/*
 * LitePCIe library
 *
 * This file is part of LitePCIe.
 *
 * Copyright (C) 2018-2020 / EnjoyDigital  / florent@enjoy-digital.fr
 * SPDX-License-Identifier: BSD-2-Clause
 */

#ifndef LITEPCIE_LIB_H
#define LITEPCIE_LIB_H

#ifdef __cplusplus
extern "C" {
#endif

#include "litepcie_dma.h"
#include "litepcie_flash.h"
#include "litepcie_helpers.h"
#include "litepcie.h"

#ifdef __cplusplus
}
#endif

#endif /* LITEPCIE_LIB_H */
